/*
 * sound creation
 */
#include <stdio.h>
#include <errno.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>
#include <xview/xview.h>
#include <xview/panel.h>
#include <multimedia/libaudio.h>
#include <multimedia/audio_device.h>
#include <multimedia/audio_hdr.h>
#include <multimedia/ulaw2linear.h>

#include "xplayer.h"
#include "xp_defs.h"

/*
 * Externs
 */

extern	void		dsp_msg();	/* xp_utils.c		*/
extern	void		start_play();	/* xp_utils.c		*/
extern	void		frm_busy();	/* xp_utils.c		*/
extern	void		xp_log();	/* xp_utils.c		*/
extern	void		speed();	/* xp_speed.c		*/

extern	int	nb_loaded;	/* nombre de sons en memoire	*/
extern	int	nb_selected;	/* nombre de sons a jouer	*/
extern	Sau_desc	au[MAX_AU];	/* loaded sounds	*/

/*
 * Globals
 */

double		tsin[SMP_RATE];	/* samples of a 1 HZ sine	*/

/*
 * initialize usin
 */

void		cr_init()
{
  int		i;
  double	inc, angle;
  
  for(angle = -PI, inc = 2*PI / (SMP_RATE-1), i = 0;
      i < SMP_RATE;
      i++, angle += inc)
  {
    tsin[i] = sin(angle);
  }
}

/*
 * create a sound
 */

void		au_create()
{
  static int	init = FALSE;
  int		freq, mag, slot;
  int		size, hpsize, i, ind;
  char		temp[120];
  unsigned char	*pdata;
  double	fact, d, dprec, valsin;
  Xv_opaque	set_snd;
  
    
  dsp_msg("Computing sound");
  frm_busy(Xplayer_pop_create->pop_create, TRUE);
  xv_set(Xplayer_win_xplayer->btn_log,
		PANEL_INACTIVE, FALSE,
		NULL);
  xv_set(Xplayer_pop_log->pop_log,
		XV_SHOW, TRUE,
		NULL);
  
  if(!init)
  {
    cr_init();
    init = TRUE;
  }
  
  /* get the sound specifications			*/
  
  freq = (int) xv_get(Xplayer_pop_create->num_frequency, PANEL_VALUE);
  if(freq > 4000)
  {
    frm_busy(Xplayer_pop_create->pop_create, FALSE);
    return;
  }
  mag = (int) xv_get(Xplayer_pop_create->num_magnitude, PANEL_VALUE);
  slot = (int) xv_get(Xplayer_pop_create->set_cslot, PANEL_VALUE);
   
  sprintf(temp, "\nCreating sound in slot %d\n", slot+1);
  xp_log(temp);
  
  size = SMP_RATE / freq;
  hpsize = SMP_RATE / (freq * 2);	/* size of a half period */
  if(size < MIN_SIZE)
    size = size * ((MIN_SIZE + size - 1)/ size);
      
  /* initialize the descriptor				*/
  
  if(au[slot].selected == E_NO_SOUND)	/* no sound in this slot*/
    nb_loaded++;
  else
  {
    if(au[slot].selected == E_ON)
      nb_selected -= 1;
      
    au[slot].selected = E_NO_SOUND;
    (void) free((char *) au[slot].data);
    /* no need to initialize the rest, it will be overwritten	*/
  }
  (au[slot].path)[0] = (au[slot].name)[0] = '\0';
  sprintf(au[slot].name, "f_%04d_m_%03d.au", freq, mag);
  sprintf(au[slot].info, "f = %04d Hz - m = %03d", freq, mag);
  au[slot].desc.sample_rate = SMP_RATE;
  au[slot].desc.encoding = AUDIO_ENCODING_ULAW;
  au[slot].desc.bytes_per_unit = 1;
  au[slot].desc.samples_per_unit = 1;
  au[slot].desc.channels = 1;
  
  au[slot].magnitude = mag / 100.0;
  
  /* allocate a buffer, shrinking if request is too big	*/
   
  size += hpsize;
  au[slot].desc.data_size = size;
  do 
  {
    au[slot].data = (unsigned char *)malloc(size);
  } while ((au[slot].data == NULL) && (size = size - (size / 8)));
    
  if(size != au[slot].desc.data_size)
  {
    /* Not enough memory to store the whole sound		*/
    sprintf(temp, "%.2f seconds of data truncated from '%s'.",
		    audio_bytes_to_secs(&au[slot].desc,
		    (size - au[slot].desc.data_size)), au[slot].name);
    xp_log(temp);
  }
  size -= hpsize;
  au[slot].desc.data_size = size;
  
  /* Create the sound						*/
  
  fact = mag/100.0;
  for(i=0, ind = 0, pdata = au[slot].data;
      i < size;
      i++, ind = (ind + freq) % SMP_RATE, pdata++)
  {
    valsin = tsin[ind];
    *pdata = audio_s2u(audio_d2s(valsin * fact));
    
  }
  
  /* this is a tricky part. We add a few more samples to the	*/
  /* created sound, so that the last sample value is as close	*/
  /* as possible to zero. Since the value of the first sample	*/
  /* of the sound is 0 (sin(-PI)), this will allows us to be	*/
  /* able to cycle on the same sound several times and get a	*/
  /* nice continuous tone					*/
  
  dprec = ABS(valsin);
  for( ;
      i < (size + hpsize); 
      i++, ind = (ind + freq) % SMP_RATE, pdata++)
  {
    valsin = tsin[ind];
    d = ABS(valsin);
    if(d < dprec)
    {
      dprec = d;
      *pdata = audio_s2u(audio_d2s(valsin * fact));
    }
    else
      break;
  }
  
  au[slot].desc.data_size = i - 1;
  
  xv_set(Xplayer_win_xplayer->btn_speed,
		PANEL_INACTIVE, FALSE,
		NULL);
    
  xv_set(Time_popup_pop_time->pop_time,
		XV_SHOW, TRUE,
		NULL);
		
  /* Display some stats of the sound file, now that it has	*/
  /* been sucessfully created					*/
  au[slot].selected = E_OFF;
  sprintf(temp, "\tInfo : %s\n", au[slot].info);
  xp_log(temp);
  au[slot].length = audio_bytes_to_secs(&au[slot].desc, 
  				au[slot].desc.data_size);
  sprintf(temp, "\tSize : %d octets\n", au[slot].desc.data_size);
  xp_log(temp);
  sprintf(temp, "\tLength : %f s\n", au[slot].length);
  xp_log(temp);
  sprintf(temp, "\tMagnitude : %f\n", au[slot].magnitude);
  xp_log(temp);
  
  /* Display the file name on the control panel		*/

  switch(slot + 1)
  {
    case 1 :
      set_snd = Xplayer_win_xplayer->set_snd1;
      break;
    
    case 2 :
      set_snd = Xplayer_win_xplayer->set_snd2;
      break;
      
    case 3 :
      set_snd = Xplayer_win_xplayer->set_snd3;
      break;
      
    case 4 :
      set_snd = Xplayer_win_xplayer->set_snd4;
      break;
      
    case 5 :
      set_snd = Xplayer_win_xplayer->set_snd5;
      break;
    
    default :
      fputs("Wrong slot\n", stderr);
      quit();
  }
  xv_set(set_snd, PANEL_INACTIVE, FALSE,
        	PANEL_CHOICE_STRING, 0, au[slot].name,
        	PANEL_VALUE, 0,
        	NULL);
        	
  if(nb_selected == 0)
    stop_play();
  else
    start_play();
      
  dsp_msg("");
  frm_busy(Xplayer_pop_create->pop_create, FALSE);

}
